/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.nephos3d.resources.programs;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.List;
import org.joml.Matrix4f;
import org.joml.Quaternionf;
import org.joml.Vector3f;
import fr.amap.lidar.amapvox.nephos3d.resources.Resource;
import fr.amap.lidar.amapvox.nephos3d.resources.materials.Material;
import fr.amap.lidar.amapvox.nephos3d.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.nephos3d.resources.shaders.Shader;
import java.util.ArrayList;
import java.util.Collections;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * The program is combined with a shader to display multiples materials.
 */
@ThreadSafe
public abstract class Program extends Resource {
    
    @GuardedBy("this")
    private String layout;
    
    @GuardedBy("this")
    private Mesh mesh;
    
    @GuardedBy("this")
    protected Shader shader;
    
    private final List<Material> materials;
    private final List<Vector3f> positions;
    private final List<Quaternionf> rotations;
    private final List<Vector3f> scales;
    
    private final Matrix4f view;
    private final Matrix4f projection;
    private final Vector3f viewPos;
    
    @GuardedBy("this")
    private int viewLocation;
    
    @GuardedBy("this")
    private int projLocation;
    
    @GuardedBy("this")
    private int viewPosLocation;
    
    @GuardedBy("this")
    private IntBuffer vbo;
    
    public Program(final String identifier) {
        super(identifier);
        this.mesh = null;
        this.materials = new ArrayList();
        this.positions = new ArrayList();
        this.rotations = new ArrayList();
        this.scales = new ArrayList();
        this.view = new Matrix4f().identity();
        this.projection = new Matrix4f().identity();
        this.viewPos = new Vector3f();
        this.vbo = Buffers.newDirectIntBuffer(3);
    }
    
    public final synchronized void addMaterial(final Material material) {
        this.materials.add(material);
    }

    public final synchronized void setMaterials(final List<Material> materials) {
        this.materials.addAll(materials);
    }
    
    public final synchronized List<Material> getMaterials() {
        return Collections.unmodifiableList(materials);
    }

    public final synchronized void setLayout(final String layout) {
        this.layout = layout;
    }

    public final synchronized String getLayout() {
        return new String(layout);
    }
    
    public final synchronized void clearMaterials() {
        this.materials.clear();
    }

    public final synchronized void setMesh(final Mesh mesh) {
        this.mesh = mesh;
    }
    
    @Override
    protected void onLoad(final GL3bc gl) {
        mesh.load(gl);
        shader.load(gl);
        shader.bind(gl);
        
        /* Generate the vertex buffer objects */
        gl.glGenBuffers(vbo.limit(), vbo);
        
        /* Generate the position buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        gl.glEnableVertexAttribArray(3);
        gl.glVertexAttribPointer(3, 3, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glVertexAttribDivisor(3, 1);
        
        /* Generate the rotation buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
        gl.glEnableVertexAttribArray(4);
        gl.glVertexAttribPointer(4, 4, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glVertexAttribDivisor(4, 1);
        
        /* Generate the scale buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(2));
        gl.glEnableVertexAttribArray(5);
        gl.glVertexAttribPointer(5, 3, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glVertexAttribDivisor(5, 1);
        
        viewPosLocation = gl.glGetUniformLocation(shader.getProgram(), "viewPos");
        viewLocation = gl.glGetUniformLocation(shader.getProgram(), "view");
        projLocation = gl.glGetUniformLocation(shader.getProgram(), "proj");
    }
    
    public void updateProperties() {
        /* Get materials values */
        positions.clear();
        rotations.clear();
        scales.clear();
           
        /* Get each position vector and rotation quaternion */        
        getMaterials().forEach((material) -> {
            positions.add(material.getPosition());
            rotations.add(material.getRotation());
            scales.add(material.getScale());
            view.set(material.getView());
            projection.set(material.getProjection());
            viewPos.set(material.getViewPosition());
        });
    }

    @Override
    protected void onBind(final GL3bc gl) {       
        /* Bind the mesh and the shader */
        mesh.bind(gl);
        shader.bind(gl);

        /* Bind the position buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        final float[] positionsValues = new float[positions.size() * 3];
        for (int i = 0; i < positions.size(); i++) {
            positionsValues[i * 3] = positions.get(i).x;
            positionsValues[i * 3 + 1] = positions.get(i).y;
            positionsValues[i * 3 + 2] = positions.get(i).z;
        }
        final FloatBuffer positionBuf = 
                Buffers.newDirectFloatBuffer(positionsValues);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, Float.BYTES * positionBuf.limit(), 
                positionBuf, GL3bc.GL_DYNAMIC_DRAW);
        
        /* Bind the rotation buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
        final float[] rotationsValues = new float[rotations.size() * 4];
        for (int i = 0; i < rotations.size(); i++) {
            rotationsValues[i * 4] = rotations.get(i).x;
            rotationsValues[i * 4 + 1] = rotations.get(i).y;
            rotationsValues[i * 4 + 2] = rotations.get(i).z;
            rotationsValues[i * 4 + 3] = rotations.get(i).w;
        }
        final FloatBuffer rotationBuf = 
                Buffers.newDirectFloatBuffer(rotationsValues);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, Float.BYTES * rotationBuf.limit(), 
                rotationBuf, GL3bc.GL_DYNAMIC_DRAW);
        
        /* Bind the scale buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(2));
        final float[] scaleValues = new float[scales.size() * 3];
        for (int i = 0; i < scales.size(); i++) {
            scaleValues[i * 3] = scales.get(i).x;
            scaleValues[i * 3 + 1] = scales.get(i).y;
            scaleValues[i * 3 + 2] = scales.get(i).z;
        }
        final FloatBuffer scaleBuf = 
                Buffers.newDirectFloatBuffer(scaleValues);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, Float.BYTES * scaleBuf.limit(), 
                scaleBuf, GL3bc.GL_DYNAMIC_DRAW);
        
        /* Send the view tranformation matrix */
        final FloatBuffer viewBuf = Buffers.newDirectFloatBuffer(16);
        gl.glUniformMatrix4fv(viewLocation, 1, false, view.get(viewBuf));
        
        /* Send the view tranformation matrix */
        final FloatBuffer projBuf = Buffers.newDirectFloatBuffer(16);
        gl.glUniformMatrix4fv(projLocation, 1, false, projection.get(projBuf));
        
        /* Send the view position vector */
        gl.glUniform3f(viewPosLocation, viewPos.x, viewPos.y, viewPos.z);
    }
    
    public final void draw(final GL3bc gl) {
        if (mesh != null) {
            mesh.setInstances(materials.size());
            mesh.draw(gl);
        }
    }

    @Override
    protected void onUnbind(final GL3bc gl) {
        shader.unbind(gl);
    }

    @Override
    protected void onUnload(final GL3bc gl) {
        shader.unload(gl);
    }
    
}
